@@ -1,9 +1,9 @@
 from __future__ import annotations
-
+from node_graph.registry import RegistryHub, registry_hub
 from node_graph.collection import NodeCollection, LinkCollection
 from uuid import uuid1
-from node_graph.sockets import SocketPool
-from node_graph.nodes import NodePool
+from node_graph.node_spec import NodeSpec
+from node_graph.socket_spec import SocketSpec, BaseSocketSpecAPI
 from typing import Dict, Any, List, Optional, Union, Callable
 import yaml
 from node_graph.node import Node
@@ -11,6 +11,7 @@ from node_graph.socket import NodeSocket
 from node_graph.link import NodeLink
 from node_graph.utils import yaml_to_dict
 from node_graph_widget import NodeGraphWidget
+from .config import BuiltinPolicy, BUILTIN_NODES
 
 
 class NodeGraph:
@@ -40,19 +41,25 @@ class NodeGraph:
         >>> ng.to_dict()
     """
 
-    NodePool: Dict[str, Any] = NodePool
-    SocketPool = SocketPool
+    registry: Optional[RegistryHub] = registry_hub
     platform: str = "node_graph"
+    _socket_spec = BaseSocketSpecAPI
+    Builtins: BuiltinPolicy = BuiltinPolicy()
 
     def __init__(
         self,
         name: str = "NodeGraph",
+        inputs: Optional[SocketSpec | List[str]] = None,
+        outputs: Optional[SocketSpec | List[str]] = None,
         uuid: Optional[str] = None,
         graph_type: str = "NORMAL",
+        graph: Optional[NodeGraph] = None,
+        parent: Optional[Node] = None,
         state: str = "CREATED",
         action: str = "NONE",
         description: str = "",
         interactive_widget: bool = False,
+        init_graph_level_nodes: bool = True,
     ) -> None:
         """Initializes a new instance of the NodeGraph class.
 
@@ -61,70 +68,124 @@ class NodeGraph:
             uuid (str, optional): The UUID of the node graph. Defaults to None.
             graph_type (str, optional): The type of the node graph. Defaults to "NORMAL".
         """
+
         self.name = name
         self.uuid = uuid or str(uuid1())
         self.graph_type = graph_type
-        self.nodes = NodeCollection(self, pool=self.NodePool)
+        self.graph = graph
+        self.parent = parent
+        self.NodePool = self.registry.node_pool
+        self.SocketPool = self.registry.socket_pool
+        self.PropertyPool = self.registry.property_pool
+        self.type_mapping = dict(self.registry.type_mapping)
+        self.nodes = NodeCollection(graph=self, pool=self.NodePool)
         self.links = LinkCollection(self)
-        self.init_meta_nodes()
         self.state = state
         self.action = action
         self.description = description
         self._widget = None
         self.interactive_widget = interactive_widget
         self._version = 0  # keep track the changes
+        self._inputs = inputs
+        self._outputs = outputs
+        if init_graph_level_nodes:
+            self._init_graph_level_nodes(inputs, outputs)
 
-    def init_meta_nodes(self) -> None:
-        self.meta_nodes = NodeCollection(self, pool=self.NodePool)
-        # add inputs and outputs nodes
-        graph_inputs = self.meta_nodes._new("any", name="graph_inputs")
-        graph_inputs.outputs._metadata.dynamic = True
-        graph_outputs = self.meta_nodes._new("any", name="graph_outputs")
-        graph_outputs.inputs._metadata.dynamic = True
-        ctx = self.meta_nodes._new("any", name="graph_ctx")
-        ctx.inputs._metadata.dynamic = True
-        ctx.inputs._default_link_limit = 1e6
+    def _init_graph_level_nodes(self, inputs, outputs, ctx=None):
+        from dataclasses import replace
+
+        inputs = self._socket_spec.validate_socket_data(inputs)
+        outputs = self._socket_spec.validate_socket_data(outputs)
+        ctx = self._socket_spec.validate_socket_data(ctx)
+
+        base_class = self.NodePool["graph_level"].load()
+
+        # graph inputs
+        inputs = self._socket_spec.dynamic(Any) if inputs is None else inputs
+        meta = replace(inputs.meta, sub_socket_default_link_limit=1000000)
+        inputs = replace(inputs, meta=meta)
+        self.graph_inputs_spec = NodeSpec(
+            identifier="graph_inputs",
+            inputs=inputs,
+            base_class=base_class,
+        )
+        self.nodes._new(self.graph_inputs_spec, name="graph_inputs")
+        # graph outputs
+        outputs = self._socket_spec.dynamic(Any) if outputs is None else outputs
+        self.graph_outputs_spec = NodeSpec(
+            identifier="graph_outputs",
+            inputs=outputs,
+            base_class=base_class,
+        )
+        graph_outputs = self.nodes._new(self.graph_outputs_spec, name="graph_outputs")
+        graph_outputs.inputs._name = "outputs"
+        # graph context
+        ctx = self._socket_spec.dynamic(Any) if ctx is None else ctx
+        meta = replace(ctx.meta, sub_socket_default_link_limit=1000000)
+        ctx = replace(ctx, meta=meta)
+        self.graph_ctx_spec = NodeSpec(
+            identifier="graph_ctx",
+            inputs=ctx,
+            base_class=base_class,
+        )
+        graph_ctx = self.nodes._new(self.graph_ctx_spec, name="graph_ctx")
+        graph_ctx.inputs._name = "ctx"
+
+    @property
+    def graph_inputs(self) -> Node:
+        """Group inputs node."""
+        return self.nodes["graph_inputs"]
+
+    @property
+    def graph_outputs(self) -> Node:
+        """Group outputs node."""
+        return self.nodes["graph_outputs"]
+
+    @property
+    def graph_ctx(self) -> Node:
+        """Context variable node."""
+        return self.nodes["graph_ctx"]
 
     @property
     def inputs(self) -> Node:
         """Group inputs node."""
-        return self.meta_nodes["graph_inputs"].outputs
+        return self.graph_inputs.inputs
 
     @inputs.setter
     def inputs(self, value: Dict[str, Any]) -> None:
         """Set group inputs node."""
-        self.meta_nodes["graph_inputs"].outputs._clear()
-        self.meta_nodes["graph_inputs"].outputs._set_socket_value(value)
+        self.graph_inputs.inputs._set_socket_value(value)
 
     @property
     def outputs(self) -> Node:
         """Group outputs node."""
-        return self.meta_nodes["graph_outputs"].inputs
+        return self.graph_outputs.inputs
 
     @outputs.setter
     def outputs(self, value: Dict[str, Any]) -> None:
         """Set group outputs node."""
-        self.meta_nodes["graph_outputs"].inputs._clear()
-        self.meta_nodes["graph_outputs"].inputs._set_socket_value(value)
+        self.graph_outputs.inputs._set_socket_value(value)
 
     @property
     def ctx(self) -> Node:
         """Context node."""
-        return self.meta_nodes["graph_ctx"].inputs
+        return self.graph_ctx.inputs
 
     @ctx.setter
     def ctx(self, value: Dict[str, Any]) -> None:
         """Set context node."""
-        self.meta_nodes["graph_ctx"].inputs._clear()
-        self.meta_nodes["graph_ctx"].inputs._set_socket_value(value, link_limit=100000)
+        self.graph_ctx.inputs._clear()
+        self.graph_ctx.inputs._set_socket_value(value, link_limit=100000)
 
     def generate_inputs(self) -> None:
         """Generate group inputs from nodes."""
         self.inputs._clear()
         for node in self.nodes:
+            if node.name in BUILTIN_NODES:
+                continue
             # skip linked sockets
             socket = node.inputs._copy(
-                node=self.meta_nodes["graph_inputs"],
+                node=self.graph_inputs,
                 parent=self.inputs,
                 skip_linked=True,
                 skip_builtin=True,
@@ -144,8 +205,10 @@ class NodeGraph:
         """Generate group outputs from nodes."""
         self.outputs._clear()
         for node in self.nodes:
+            if node.name in BUILTIN_NODES:
+                continue
             socket = node.outputs._copy(
-                node=self.meta_nodes["graph_outputs"],
+                node=self.graph_outputs,
                 parent=self.outputs,
                 skip_builtin=True,
             )
@@ -160,54 +223,14 @@ class NodeGraph:
                 # add link from node outputs to group outputs
                 self.add_link(node.outputs[key], self.outputs[new_key])
 
-    @property
-    def meta_sockets(self) -> Node:
-        """Meta sockets node."""
-        return {
-            "graph_inputs": self.inputs,
-            "graph_outputs": self.outputs,
-            "graph_ctx": self.ctx,
-        }
-
-    def meta_sockets_to_dict(self) -> Dict[str, Any]:
-        meta_sockets = {
-            "graph_ctx": self.ctx._to_dict(),
-            "graph_inputs": self.inputs._to_dict(),
-            "graph_outputs": self.outputs._to_dict(),
-        }
-        return meta_sockets
-
-    def meta_sockets_from_dict(self, meta_sockets: Dict[str, Any]) -> None:
-        ctx_data = meta_sockets.get("graph_ctx", {})
-        if ctx_data:
-            self.meta_nodes["graph_ctx"].inputs = self.ctx.__class__._from_dict(
-                ctx_data,
-                node=self.meta_nodes["graph_ctx"],
-                pool=self.SocketPool,
-                graph=self,
-            )
-        else:
-            self.meta_nodes["graph_ctx"].inputs._clear()
-        inputs_data = meta_sockets.get("graph_inputs", {})
-        if inputs_data:
-            self.meta_nodes["graph_inputs"].outputs = self.inputs.__class__._from_dict(
-                inputs_data,
-                node=self.meta_nodes["graph_inputs"],
-                pool=self.SocketPool,
-                graph=self,
-            )
-        else:
-            self.meta_nodes["graph_inputs"].outputs._clear()
-        outputs_data = meta_sockets.get("graph_outputs", {})
-        if outputs_data:
-            self.meta_nodes["graph_outputs"].inputs = self.outputs.__class__._from_dict(
-                outputs_data,
-                node=self.meta_nodes["graph_outputs"],
-                pool=self.SocketPool,
-                graph=self,
-            )
-        else:
-            self.meta_nodes["graph_outputs"].inputs._clear()
+    def set_inputs(self, inputs: Dict[str, Any]):
+        for name, input in inputs.items():
+            if "graph_inputs" in self.nodes and name in self.inputs:
+                setattr(self.inputs, name, input)
+            elif name in self.nodes:
+                self.nodes[name].set_inputs(input)
+            else:
+                raise KeyError(f"{name} not found in WorkGraph inputs or nodes.")
 
     @property
     def platform_version(self) -> str:
@@ -227,32 +250,51 @@ class NodeGraph:
         return self._widget
 
     def add_node(
-        self, identifier: Union[str, Callable], name: str = None, **kwargs
+        self,
+        identifier: Union[str, Callable],
+        name: str = None,
+        include_builtins: bool = False,
+        **kwargs,
     ) -> Node:
         """Adds a node to the node graph."""
 
         from node_graph.decorator import build_node_from_callable
-        from node_graph.nodes.factory.nodegraph_node import NodeGraphNodeFactory
+        from node_graph.node_spec import NodeHandle
+        from node_graph.nodes.subgraph_node import _build_subgraph_task_nodespec
 
-        if name in ["graph_ctx", "graph_inputs", "graph_inputs"]:
+        if name in BUILTIN_NODES and not include_builtins:
             raise ValueError(f"Name {name} can not be used, it is reserved.")
 
         if isinstance(identifier, NodeGraph):
-            identifier = NodeGraphNodeFactory.create_node(identifier)
+            identifier = _build_subgraph_task_nodespec(graph=identifier, name=name)
         # build the task on the fly if the identifier is a callable
-        elif callable(identifier):
-            if hasattr(identifier, "_NodeCls"):
-                identifier = identifier._NodeCls
-            else:
-                identifier = build_node_from_callable(identifier)
+        elif callable(identifier) and not isinstance(
+            identifier, (NodeSpec, NodeHandle)
+        ):
+            identifier = build_node_from_callable(identifier)
         node = self.nodes._new(identifier, name, **kwargs)
         self._version += 1
         return node
 
     def add_link(self, source: NodeSocket | Node, target: NodeSocket) -> NodeLink:
         """Add a link between two nodes."""
+        from node_graph.socket import NodeSocketNamespace
+
         if isinstance(source, Node):
             source = source.outputs["graph_outputs"]
+        elif source._parent is None and isinstance(source, NodeSocketNamespace):
+            # if the source is the top-level outputs,
+            # we use the built-in "_outputs" socket to represent it
+            if "_outputs" in source:
+                source = source["_outputs"]
+            else:
+                raise ValueError(
+                    f"You try to link a top-level output socket {source._name} without a parent."
+                )
+        #
+        key = f"{source._node.name}.{source._scoped_name} -> {target._node.name}.{target._scoped_name}"
+        if key in self.links:
+            return self.links[key]
         link = self.links._new(source, target)
         self._version += 1
         return link
@@ -307,7 +349,7 @@ class NodeGraph:
             short=short, should_serialize=should_serialize
         )
 
-        links, meta_links = self.links_to_dict()
+        links = self.links_to_dict()
         data = {
             "platform_version": f"{self.platform}@{self.platform_version}",
             "uuid": self.uuid,
@@ -317,25 +359,33 @@ class NodeGraph:
             "error": "",
             "metadata": metadata,
             "nodes": nodes,
-            "meta_sockets": self.meta_sockets_to_dict(),
             "links": links,
-            "meta_links": meta_links,
             "description": self.description,
         }
         return data
 
     def get_metadata(self) -> Dict[str, Any]:
-        """Converts the metadata to a dictionary.
-
-        Returns:
-            Dict[str, Any]: The metadata.
+        """Export graph metadata including *live* graph-level IO specs.
+        We snapshot current shapes of graph inputs/outputs/ctx using `SocketSpec.from_namespace`.
         """
-        metadata: Dict[str, Any] = {
+        meta: Dict[str, Any] = {
             "graph_type": self.graph_type,
-            # "inputs": self.inputs,
-            # "outputs": self.outputs,
+            "inputs_spec": SocketSpec.from_namespace(
+                self.graph_inputs.inputs, type_mapping=self.type_mapping
+            ).to_dict(),
+            "outputs_spec": SocketSpec.from_namespace(
+                self.graph_outputs.inputs, type_mapping=self.type_mapping
+            ).to_dict(),
+            "ctx_spec": SocketSpec.from_namespace(
+                self.graph_ctx.inputs, type_mapping=self.type_mapping
+            ).to_dict(),
         }
-        return metadata
+        # also save the parent class information
+        meta["graph_class"] = {
+            "callable_name": self.__class__.__name__,
+            "module_path": self.__class__.__module__,
+        }
+        return meta
 
     def export_nodes_to_dict(
         self, short: bool = False, should_serialize: bool = False
@@ -348,6 +398,7 @@ class NodeGraph:
         Returns:
             Dict[str, Any]: The nodes data.
         """
+        # generate spec for graph-level nodes
         nodes = {}
         for node in self.nodes:
             nodes[node.name] = node.to_dict(
@@ -362,16 +413,9 @@ class NodeGraph:
             List[Dict[str, Any]]: The links data.
         """
         links = []
-        meta_links = []
         for link in self.links:
-            if (
-                link.from_node.name in self.meta_nodes
-                or link.to_node.name in self.meta_nodes
-            ):
-                meta_links.append(link.to_dict())
-            else:
-                links.append(link.to_dict())
-        return links, meta_links
+            links.append(link.to_dict())
+        return links
 
     def to_yaml(self) -> str:
         """Exports the node graph to a YAML format data.
@@ -390,7 +434,7 @@ class NodeGraph:
         """Updates the node graph from the database."""
         raise NotImplementedError("The 'update' method is not implemented.")
 
-    def links_from_dict(self, links: list, meta_links: list) -> None:
+    def links_from_dict(self, links: list) -> None:
         """Adds links to the node graph from a dictionary.
 
         Args:
@@ -401,42 +445,9 @@ class NodeGraph:
                 self.nodes[link["from_node"]].outputs[link["from_socket"]],
                 self.nodes[link["to_node"]].inputs[link["to_socket"]],
             )
-        # add meta links
-        for link in meta_links:
-            if link["from_node"] in self.meta_sockets:
-                meta_socket = self.meta_sockets[link["from_node"]]
-                if link["from_socket"] not in meta_socket:
-                    meta_socket._set_socket_value(
-                        {link["from_socket"]: None}, link_limit=100000
-                    )
-                if link["to_node"] in self.meta_sockets:
-                    if link["to_socket"] not in self.meta_sockets[link["to_node"]]:
-                        self.meta_sockets[link["to_node"]]._set_socket_value(
-                            {link["to_socket"]: None}, link_limit=100000
-                        )
-                    to_socket = self.meta_sockets[link["to_node"]][link["to_socket"]]
-                else:
-                    to_socket = self.nodes[link["to_node"]].inputs[link["to_socket"]]
-                self.add_link(meta_socket[link["from_socket"]], to_socket)
-            elif link["to_node"] in self.meta_sockets:
-                meta_socket = self.meta_sockets[link["to_node"]]
-                if link["from_node"] in self.meta_sockets:
-                    from_socket = self.meta_sockets[link["from_node"]][
-                        link["from_socket"]
-                    ]
-                else:
-                    from_socket = self.nodes[link["from_node"]].outputs[
-                        link["from_socket"]
-                    ]
-                meta_socket._set_socket_value(
-                    {link["to_socket"]: from_socket},
-                    link_limit=100000,
-                )
 
     @classmethod
-    def from_dict(
-        cls, ngdata: Dict[str, Any], class_factory: Callable = None
-    ) -> "NodeGraph":
+    def from_dict(cls, ngdata: Dict[str, Any]) -> "NodeGraph":
         """Rebuilds a node graph from a dictionary.
 
         Args:
@@ -445,10 +456,9 @@ class NodeGraph:
         Returns:
             NodeGraph: The rebuilt node graph.
         """
-        if class_factory is None:
-            from node_graph.nodes.factory.base import BaseNodeFactory
 
-            class_factory = BaseNodeFactory
+        md = ngdata.get("metadata", {}) or {}
+
         ng: "NodeGraph" = cls(
             name=ngdata["name"],
             uuid=ngdata.get("uuid"),
@@ -456,23 +466,69 @@ class NodeGraph:
             state=ngdata.get("state", "CREATED"),
             action=ngdata.get("action", "NONE"),
             description=ngdata.get("description", ""),
+            init_graph_level_nodes=False,
         )
-        for name, ndata in ngdata["nodes"].items():
-            if ndata.get("metadata", {}).get("is_dynamic", False):
-                identifier = class_factory(ndata)
+        # built-in graph level nodes
+        inputs_spec = (
+            SocketSpec.from_dict(md["inputs_spec"], type_mapping=ng.type_mapping)
+            if "inputs_spec" in md
+            else None
+        )
+        outputs_spec = (
+            SocketSpec.from_dict(md["outputs_spec"], type_mapping=ng.type_mapping)
+            if "outputs_spec" in md
+            else None
+        )
+        ctx_spec = (
+            SocketSpec.from_dict(md["ctx_spec"], type_mapping=ng.type_mapping)
+            if "ctx_spec" in md
+            else None
+        )
+        ng._init_graph_level_nodes(inputs_spec, outputs_spec, ctx_spec)
+
+        for ndata in ngdata["nodes"].values():
+            ng.add_node_from_dict(ndata)
+
+        ng.links_from_dict(ngdata.get("links", []))
+        return ng
+
+    def add_node_from_dict(self, ndata: Dict[str, Any]) -> Node:
+        """Adds a node to the node graph from a dictionary.
+
+        Args:
+            ndata (Dict[str, Any]): The data of the node.
+
+        Returns:
+            Node: The added node.
+        """
+        import importlib
+
+        md = ndata.get("metadata", {}) or {}
+        name = ndata["name"]
+        if name in BUILTIN_NODES:
+            self.nodes[name].update_from_dict(ndata)
+            return self.nodes[name]
+        else:
+            if md.get("is_dynamic", False):
+                node_class = ndata.get("metadata", {}).get("node_class", None)
+                node_class = getattr(
+                    importlib.import_module(node_class["module_path"]),
+                    node_class["callable_name"],
+                )
+                node = node_class.from_dict(ndata, graph=self)
+                self.nodes._append(node)
             else:
                 identifier = ndata["identifier"]
-            node = ng.add_node(
-                identifier,
-                name=name,
-                uuid=ndata.pop("uuid", None),
-                _metadata=ndata.get("metadata", None),
-                _executor=ndata.get("executor", None),
-            )
-            node.update_from_dict(ndata)
-        ng.meta_sockets_from_dict(ngdata.get("meta_sockets", {}))
-        ng.links_from_dict(ngdata.get("links", []), ngdata.get("meta_links", []))
-        return ng
+                node = self.add_node(
+                    identifier,
+                    name=name,
+                    uuid=ndata.pop("uuid", None),
+                    _metadata=ndata.get("metadata", None),
+                    _executor=ndata.get("executor", None),
+                    include_builtins=ndata["name"] in BUILTIN_NODES,
+                )
+                node.update_from_dict(ndata)
+        return node
 
     @classmethod
     def from_yaml(
@@ -587,7 +643,11 @@ class NodeGraph:
         Returns:
             NodeGraph: The combined node graph.
         """
-        self.nodes._extend(other.nodes._copy(graph=self))
+        nodes = other.nodes._copy(graph=self)
+        for node in nodes:
+            # skip built-in nodes
+            if node.name not in BUILTIN_NODES:
+                self.nodes._append(node)
         for link in other.links:
             self.add_link(
                 self.nodes[link.from_node.name].outputs[link.from_socket._name],
@@ -649,3 +709,13 @@ class NodeGraph:
         """Write a standalone html file to visualize the graph."""
         self.widget.value = self.to_widget_value()
         return self.widget.to_html(output=output, **kwargs)
+
+    def __enter__(self):
+        from node_graph.manager import active_graph as _active_graph
+
+        self.___ctx = _active_graph(self)
+        self.___ctx.__enter__()
+        return self
+
+    def __exit__(self, exc_type, exc, tb):
+        return self.___ctx.__exit__(exc_type, exc, tb)
