@@ -1,4 +1,4 @@
-from node_graph.collection import EntryPointPool
+from node_graph.registry import EntryPointPool
 
 # global instance
 PropertyPool = EntryPointPool(entry_point_group="node_graph.property")
