@@ -1,10 +1,12 @@
 from __future__ import annotations
-
-from node_graph.group import DependencyGroup
+from node_graph.collection import DependencyCollection
 from node_graph.property import NodeProperty
 from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union
-from node_graph.collection import get_item_class, EntryPointPool
-from dataclasses import dataclass, field, asdict
+from node_graph.collection import get_item_class
+from dataclasses import dataclass, field
+from node_graph.orm.mapping import type_mapping
+from node_graph.registry import EntryPointPool
+import wrapt
 
 if TYPE_CHECKING:
     from node_graph.node import Node
@@ -83,7 +85,7 @@ class OperatorSocketMixin:
             raise ValueError("Socket does not belong to a WorkGraph.")
 
         new_node = graph.nodes._new(
-            self._decorator()(op_func)._NodeCls,
+            self._decorator()(op_func),
             x=x,
             y=y,
         )
@@ -155,24 +157,24 @@ class OperatorSocketMixin:
     def __ne__(self, other):
         return self._create_operator_node(op_ne, self, other)
 
-    def __rshift__(self, other: BaseSocket | Node | DependencyGroup):
+    def __rshift__(self, other: BaseSocket | Node | DependencyCollection):
         """
         Called when we do: self >> other
         So we link them or mark that 'other' must wait for 'self'.
         """
-        if isinstance(other, DependencyGroup):
+        if isinstance(other, DependencyCollection):
             for item in other.items:
                 self >> item
         else:
             other._waiting_on.add(self)
         return other
 
-    def __lshift__(self, other: BaseSocket | Node | DependencyGroup):
+    def __lshift__(self, other: BaseSocket | Node | DependencyCollection):
         """
         Called when we do: self << other
         Means the same as: other >> self
         """
-        if isinstance(other, DependencyGroup):
+        if isinstance(other, DependencyCollection):
             for item in other.items:
                 self << item
         else:
@@ -225,6 +227,7 @@ class SocketMetadata:
     """
 
     dynamic: bool = False
+    sub_socket_default_link_limit: int = 1
     required: bool = False
     builtin_socket: bool = False
     function_socket: bool = False
@@ -235,7 +238,18 @@ class SocketMetadata:
     def to_dict(self) -> Dict[str, Any]:
         """Return a *plain* dict suitable for JSON serialisation."""
 
-        data = asdict(self)
+        data = {
+            "dynamic": self.dynamic,
+            "required": self.required,
+            "builtin_socket": self.builtin_socket,
+            "function_socket": self.function_socket,
+            "socket_type": self.socket_type,
+            "arg_type": self.arg_type,
+        }
+        if self.extras:
+            # Make sure extras are a dict
+            data["extras"] = dict(self.extras)
+
         # Do not bloat output with empty *extras*
         if not data.get("extras"):
             data.pop("extras", None)
@@ -260,6 +274,7 @@ class SocketMetadata:
                 "socket_type",
                 "arg_type",
                 "required",
+                "sub_socket_default_link_limit",
                 "extras",
             }
             known = {k: v for k, v in raw.items() if k in known_keys}
@@ -273,6 +288,50 @@ class SocketMetadata:
         )
 
 
+class TaggedValue(wrapt.ObjectProxy):
+    def __init__(self, wrapped, socket=None):
+        super().__init__(wrapped)
+
+        self._self_socket = socket
+
+    # Provide clean access via `proxy._socket` instead of `proxy._self_socket`
+    @property
+    def _socket(self):
+        return self._self_socket
+
+    @_socket.setter
+    def _socket(self, value):
+        self._self_socket = value
+
+    def __copy__(self):
+        # shallow-copy the wrapped value, preserve the tag
+        from copy import copy as _copy
+
+        w = _copy(self.__wrapped__)
+        return type(self)(w, socket=self._socket)
+
+    def __deepcopy__(self, memo):
+        from copy import deepcopy as _deepcopy
+
+        # required by wrapt.ObjectProxy
+        oid = id(self)
+        if oid in memo:
+            return memo[oid]
+        w = _deepcopy(self.__wrapped__, memo)
+        clone = type(self)(w, socket=self._socket)
+        memo[oid] = clone
+        return clone
+
+    def __reduce_ex__(self, protocol):
+        """
+        This is the magic method for serialization.
+
+        Instead of returning instructions to rebuild the TaggedValue proxy,
+        only the underlying value (self.__wrapped__) gets saved.
+        """
+        return self.__wrapped__.__reduce_ex__(protocol)
+
+
 class BaseSocket:
     """Socket object for input and output sockets of a Node.
 
@@ -432,7 +491,15 @@ class NodeSocket(BaseSocket, OperatorSocketMixin):
 
     def _set_socket_value(self, value: Any) -> None:
         if isinstance(value, BaseSocket):
+            if (
+                isinstance(value, NodeSocketNamespace)
+                and value._parent is None
+                and "_outputs" in value
+            ):
+                value = value._outputs
             self._node.graph.add_link(value, self)
+        elif isinstance(value, TaggedValue):
+            self._node.graph.add_link(value._socket, self)
         elif self.property:
             self.property.value = value
         else:
@@ -509,7 +576,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
     """A NodeSocket that also acts as a namespace (collection) of other sockets."""
 
     _identifier: str = "node_graph.namespace"
-    _default_link_limit = 1
+    _type_mapping: dict = type_mapping
 
     _RESERVED_NAMES = {
         "_RESERVED_NAMES",
@@ -528,7 +595,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
         name: str,
         node: Optional["Node"] = None,
         parent: Optional["NodeSocket"] = None,
-        link_limit: int = 1e6,
+        link_limit: int = 1000000,
         metadata: Union[SocketMetadata, Dict[str, Any], None] = None,
         sockets: Optional[Dict[str, object]] = None,
         pool: Optional[object] = None,
@@ -560,7 +627,7 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             self._SocketPool = SocketPool
 
         if self._metadata.dynamic:
-            self._link_limit = 1e6
+            self._link_limit = 1000000
         if sockets is not None:
             for key, socket in sockets.items():
                 kwargs = {}
@@ -600,20 +667,37 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             object.__setattr__(self, name, value)
             return
 
-        self._set_socket_value({name: value}, link_limit=self._default_link_limit)
+        self._set_socket_value(
+            {name: value}, link_limit=self._metadata.sub_socket_default_link_limit
+        )
+
+    def __setitem__(self, key: str | int, value: Any) -> None:
+        """
+        Override __setitem__ so that doing `namespace_socket[key] = x`
+        either sets the property or links to another socket, rather than
+        replacing the entire sub-socket object.
+        """
+        if isinstance(key, int):
+            key = list(self._sockets.keys())[key]
+        self.__setattr__(key, value)
 
     def __dir__(self) -> list[str]:
         """
         Make tab-completion more friendly:
         """
-        socket_attrs = set(self._sockets.keys())
-        return socket_attrs
+        # Get the list of default attributes from the parent class
+        default_attrs = super().__dir__()
+
+        # Get the custom attributes from the _sockets dictionary
+        socket_attrs = self._sockets.keys()
+
+        # Combine the default and custom attributes, remove duplicates, and sort
+        return sorted(list(set(default_attrs) | set(socket_attrs)))
 
     def _new(
         self,
         identifier: Union[str, type] = None,
         name: Optional[str] = None,
-        link_limit: int = 1,
         metadata: Optional[dict] = None,
         **kwargs: Any,
     ) -> object:
@@ -630,7 +714,10 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                     self._new(
                         self._SocketPool["namespace"],
                         namespace,
-                        metadata={"dynamic": True},
+                        metadata={
+                            "dynamic": True,
+                            "sub_socket_default_link_limit": self._metadata.sub_socket_default_link_limit,
+                        },
                     )
                 else:
                     raise ValueError(
@@ -639,17 +726,20 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             return self[namespace]._new(
                 identifier,
                 _names[1],
+                link_limit=self._metadata.sub_socket_default_link_limit,
                 metadata=metadata,
             )
         else:
-            ItemClass = get_item_class(identifier, self._SocketPool, BaseSocket)
+            ItemClass = get_item_class(identifier, self._SocketPool)
             kwargs.pop("graph", None)
+            kwargs.setdefault(
+                "link_limit", self._metadata.sub_socket_default_link_limit
+            )
             item = ItemClass(
                 name,
                 node=self._node,
                 parent=self,
                 graph=self._graph,
-                link_limit=link_limit,
                 metadata=metadata,
                 pool=self._SocketPool,
                 **kwargs,
@@ -698,7 +788,8 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                             self._new(self._SocketPool["any"], key, **kwargs)
                     else:
                         raise ValueError(
-                            f"Socket: {key} does not exist in the namespace socket: {self._name}."
+                            f"Socket: {key} does not exist in a non-dynamic namespace socket: {self._name}. "
+                            f"This namespace belongs to: {self._node.name}"
                         )
                 if isinstance(self[key], NodeSocketNamespace):
                     self[key]._set_socket_value(val, **kwargs)
@@ -747,10 +838,118 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
             **kwargs,
         )
         # Add nested sockets
-        for item_data in data.get("sockets", {}).values():
+        for name, item_data in data.get("sockets", {}).items():
+            item_data["name"] = name
             ns._new(**item_data)
         return ns
 
+    @classmethod
+    def _from_spec(
+        cls,
+        name: str,
+        spec: "SocketSpec",
+        *,
+        node: Optional["Node"],
+        graph: Optional["NodeGraph"],
+        parent: Optional["NodeSocket"] = None,
+        pool: Optional[object] = None,
+        role: str = "input",
+    ) -> "NodeSocketNamespace":
+        """
+        Materialize a runtime namespace (and children) from a SocketSpec.
+        If *spec* is a leaf (non-namespace), a single leaf socket named 'result' is created.
+        """
+        from node_graph.materialize import runtime_meta_from_spec
+
+        # check if spec.identifier is not a namespace
+        if spec.identifier != cls._type_mapping["namespace"]:
+            raise ValueError(
+                f"The socket spec identifier must be a namespace, got: {spec.identifier}"
+            )
+
+        ns_meta = runtime_meta_from_spec(
+            spec, role=role, function_generated=True, type_mapping=cls._type_mapping
+        )
+        ns = cls(
+            name=name,
+            node=node,
+            parent=parent,
+            graph=graph,
+            metadata=ns_meta,
+            pool=pool or (node.SocketPool if node else None),
+        )
+
+        # Namespace
+        parent_defaults = dict(getattr(spec, "defaults", {}) or {})
+        for fname, f_spec in (spec.fields or {}).items():
+            cls._append_from_spec(
+                ns, fname, f_spec, parent_defaults, node=node, graph=graph, role=role
+            )
+        return ns
+
+    @classmethod
+    def _append_from_spec(
+        cls,
+        parent_ns: "NodeSocketNamespace",
+        name: str,
+        spec: "SocketSpec",
+        parent_defaults: dict,
+        *,
+        node,
+        graph,
+        role: str,
+    ) -> None:
+        from node_graph.materialize import runtime_meta_from_spec
+
+        if spec.identifier == cls._type_mapping["namespace"]:
+            child_meta = runtime_meta_from_spec(
+                spec, role=role, function_generated=True, type_mapping=cls._type_mapping
+            )
+            child = cls(
+                name=name,
+                node=node,
+                parent=parent_ns,
+                graph=graph,
+                metadata=child_meta,
+                pool=parent_ns._SocketPool,
+            )
+            parent_ns._append(child)
+            # cascade defaults into nested namespace if provided as dict
+            child_defaults = (
+                parent_defaults.get(name)
+                if isinstance(parent_defaults.get(name), dict)
+                else {}
+            )
+            for fname, f_spec in (spec.fields or {}).items():
+                cls._append_from_spec(
+                    child,
+                    fname,
+                    f_spec,
+                    child_defaults,
+                    node=node,
+                    graph=graph,
+                    role=role,
+                )
+        else:
+            # leaf
+            leaf_meta = runtime_meta_from_spec(
+                spec, role=role, function_generated=True, type_mapping=cls._type_mapping
+            )
+            prop = {"identifier": spec.identifier}
+            if name in parent_defaults:
+                prop["default"] = parent_defaults[name]
+            ItemClass = get_item_class(spec.identifier, parent_ns._SocketPool)
+            sock = ItemClass(
+                name=name,
+                node=node,
+                parent=parent_ns,
+                graph=graph,
+                metadata=leaf_meta,
+                property=prop,
+                link_limit=parent_ns._metadata.sub_socket_default_link_limit,
+            )
+            parent_ns._append(sock)
+
     def _copy(
         self,
         node: Optional[Node] = None,
@@ -793,9 +992,12 @@ class NodeSocketNamespace(BaseSocket, OperatorSocketMixin):
                 if len(keys) > 1:
                     return item[keys[1]]
                 return item
+            parent_name = (
+                self._parent._name if self._parent is not None else self._node.name
+            )
             raise AttributeError(
-                f""""{key}" is not in the {self.__class__.__name__}.
-Acceptable names are {self._get_keys()}. This collection belongs to {self._parent}."""
+                f""""{key}" is not in this namespace. Acceptable names are {self._get_keys()}."""
+                f"""This namespace belongs to {parent_name}."""
             )
 
     def __contains__(self, name: str) -> bool:
